;----------------------------------------------------------------------------
; entry.asm: this code is called by the multiboot loader (GRUB for example)
;----------------------------------------------------------------------------

; At the beginning of _start we assume that:
;   * CS has read/exec attributs, start= 0, size= 4G
;   * DS, ES, FS, GS, SS have read/write attributs, start= 0, size= 4G
;   * A20 gate is useable for linear 32-bit addressing
;   * paging if off
;   * interrupt flag if off
;   * eax = 0x2BAD002 (which indicated the kernel has been loaded by a
;     multiboot-compliant loader)
;   * ebx points to a multiboot_info structure (see multiboot.h)
;
; (This context is defined in the multiboot standard 0.6 (boot-proposal.html))
;

%define STACK_SIZE 2*1024
%define KERNEL_CS 0x08
%define KERNEL_DS 0x10

[SEGMENT .text]
[BITS 32]

extern c_boot_entry
extern _edata, _end

[GLOBAL start]
start:
      ; first we jump over the boot structure (used by the loader)
      jmp .over_boot_struct

; constants defined in multiboot.h
%define MULTIBOOT_MAGIC 0x1BADB002
%define MULTIBOOT_FLAGS 0x00000003
align 4
        dd MULTIBOOT_MAGIC
        dd MULTIBOOT_FLAGS
        dd 0 - (MULTIBOOT_MAGIC + MULTIBOOT_FLAGS)

.over_boot_struct:
        
        ; now, we save the address of the multiboot_info struct
        mov dword [multiboot_info_addr], ebx
        mov dword [bootloader_signature], eax

        ; before all, we clear BSS section (uninitialized data)
        mov edi, _edata              ; destination addres
        mov ecx, _end                ; end of data
        sub ecx, edi                 ; dest-end = length
        xor eax, eax
        cld                          ; clear forward
        rep stosb                    ; do it

        ; reset EFLAGS to a known state
        push dword 0
        popfd
       
        ; setup a new GDT
        call setup_gdt

        
        ; setup the stack (SS should contain good descriptor after setup_gdt)
        mov esp, top_of_kernel_stack

        ; reprogram the PIC (programmable interrupt controller)
        call setup_pic

        ; change the timer frequency (to 100 Hz)
        call setup_pit

        ; setup a new IDT
        call setup_idt

        ; now we can enable interrupts
        
        sti

        ; let's call the "real" kernel:
        ; load the arguments
        mov eax, [multiboot_info_addr]
        mov ebx, [bootloader_signature]
        push ebx
        push eax
        call c_boot_entry

        ; just to be sure!
        hlt
die:
        jmp die

;----------------------------------------------------------------------------
extern gdtr
setup_gdt:
        lgdt [gdtr]
        
        jmp KERNEL_CS:.reload_CS        ; reload CS
        nop
.reload_CS:
        mov eax, KERNEL_DS              ; and DS, ES, FS, GS and SS
        mov ds, ax
        mov es, ax
        mov fs, ax
        mov gs, ax
        mov ss, ax
        
        ret
       
;----------------------------------------------------------------------------
extern init_intr
extern idtr
setup_idt:
        call init_intr
        lidt [idtr]
        ret
       
;----------------------------------------------------------------------------
setup_pic:
;* this code comes from John S. Fine  <johnfine@erols.com>

; defines for PIC programming:
M_PIC	EQU	0x20		;I/O for master PIC
M_IMR	EQU	0x21		;I/O for master IMR
S_PIC	EQU	0xA0		;I/O for slave PIC
S_IMR	EQU	0xA1		;I/O for slace IMR

EOI	EQU	0x20		;EOI command

ICW1	EQU	0x11		;Cascade, Edge triggered
				;ICW2 is vector
				;ICW3 is slave bitmap or number
ICW4	EQU	0x01		;8088 mode

M_VEC	EQU	0x20		;Vector for master
S_VEC	EQU	0x28		;Vector for slave

OCW3_IRR EQU	0x0A		;Read IRR
OCW3_ISR EQU	0x0B		;Read ISR

	mov	al, ICW1		;Start 8259 initialization
	out	M_PIC, al
	out	S_PIC, al
	mov	al, M_VEC		;Base interrupt vector
	out	M_PIC+1, al
	mov	al, S_VEC
	out	S_PIC+1, al
	mov	al, 1<<2		;Bitmask for cascade on IRQ 2
	out	M_PIC+1, al
	mov	al, 2			;Cascade on IRQ 2
	out	S_PIC+1, al
	mov	al, ICW4		;Finish 8259 initialization
	out	M_PIC+1, al
	out	S_PIC+1, al

	mov	al, 0xFF		;Mask all interrupts
	out	M_IMR, al
        out	S_IMR, al
        
        ret

;----------------------------------------------------------------------------
setup_pit:
        ; now, set the PIT to 100 hz
        mov al, 0x36
        out 0x43, al

        mov bx, 1193180/100     ; 100 hz
        mov al, bl
        out 0x40, al
        mov al, bh
        out 0x40, al

        ret

;----------------------------------------------------------------------------
[SEGMENT .data]

; we can't put these two variables in bss because we need they before
; we clear the BSS section.
multiboot_info_addr dd 0
bootloader_signature dd 0

;----------------------------------------------------------------------------
[SEGMENT .bss]

alignb 16
kernel_stack resb STACK_SIZE
top_of_kernel_stack

