/*
 * LittleOS - http://come.to/littleos/
 * Copyright (C) 1998-2000 Lacroix Pascal <placr@hotmail.com>
 * All rights reserved.
 *
 * LittleOS is free software -- also known as "open source" -- you can
 * redistribute it and/or modify it under the terms of the GNU General
 * Public License, version 2, as published by the Free Software Foundation,
 * and found in the file `COPYING'.  This license basically allows free use,
 * modification, and redistribution, as long as the source to LittleOS and
 * any code linked against it is made freely available.  (The license in
 * `COPYING' is authoritative, not this interpretation of it.)
 *
 * LittleOS is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with LittleOS; see the file COPYING.  If not, write to the Free Software
 * Foundation, 59 Temple Place #330, Boston, MA 02111-1307, USA, or look
 * at http://www.fsf.org/copyleft/gpl.html.
 *
 */

#include <types.h>
#include <i386/gdt.h>
#include <i386/pio.h>
#include <i386/selectors.h>
#include <sys/kernel.h>
#include <sys/intr.h>
#include <sys/syscalls.h>

#include <sys/paging.h>

/* interrupt descriptor table */
descTable(idt, 256)
{
};

/* IDT register structure used for "lidt" instruction */
struct
{
    unsigned short limit __attribute__((packed));
    union DTEntry *idt __attribute__((packed));
}
idtr =
{
    (256 * sizeof(union DTEntry) - 1), idt
};


#define SLAVE_IRQ 8
#define MASTER_SLAVE 2

#define	ICU0 0x20
#define	ICU1 0xa0

#define	ICU_RESET 0x11

static unsigned short irq_mask = 0xFFFF;

/* enables irq irq_no */
void
enable_irq(unsigned short irq_no)
{
    
	irq_mask &= ~(1 << irq_no);
    if (irq_no >= SLAVE_IRQ)
	irq_mask &= ~(1 << MASTER_SLAVE);

    outb(ICU0 + 1, irq_mask & 0xFF);
    outb(ICU1 + 1, (irq_mask >> 8) & 0xFF);

}


/* disables irq irq_no */ 
void
disable_irq(unsigned short irq_no)
{
    irq_mask |= (1 << irq_no);
    if (irq_no >= SLAVE_IRQ)
	irq_mask |= (1 << MASTER_SLAVE);
    
	outb(ICU0 + 1, irq_mask & 0xFF);
    outb(ICU1 + 1, (irq_mask >> 8) & 0xFF);
    
}

extern void dummy_int();	/* in low-level.asm */

#define INTHANDLER(name, no) extern void __int_##name();
#define INTERRHANDLER(name, no) INTHANDLER(name, no)
#define IRQAHANDLER(name, no) INTHANDLER(name, no)
#define IRQBHANDLER(name, no) INTHANDLER(name, no)
#include "interrupts.lst"
#undef IRQAHANDLER
#undef IRQBHANDLER
#undef INTERRHANDLER
#undef INTHANDLER

void
setup_idt_entry(int no, long routine, int access)
{
    idt[no].gate.offset_low = (unsigned short) (routine);
    idt[no].gate.offset_high = (unsigned short) ((long) routine >> 16);
    idt[no].gate.selector = KERNEL_CS;
    idt[no].gate.access = access;
}

/* setup interrupt table */
void
init_intr(void)
{
    int i;

    for (i = 0; i < 256; i++)	/* put a default routine for all interrupts */
	setup_idt_entry(i, (long) &dummy_int, D_INT + D_PRESENT + D_DPL3);

#define INTHANDLER(name, no) setup_idt_entry(no, (long) &__int_##name, \
    D_TRAP+D_PRESENT+D_DPL3);
#define INTERRHANDLER INTHANDLER
#define IRQAHANDLER(name, no) setup_idt_entry(no, (long) &__int_##name, \
    D_INT+D_PRESENT+D_DPL3);
#define IRQBHANDLER IRQAHANDLER
#include "interrupts.lst"
#undef IRQAHANDLER
#undef IRQBHANDLER
#undef INTERRHANDLER
#undef INTHANDLER
    __asm__ __volatile__(
			    "pushfl                   \n\t"
			    "andl $0xffffbfff, (%esp) \n\t"	/* Clear NT flag */
			    "popfl                    \n\t");

}

long int_no;			/* used in low-level.asm to pass the interrupt number */

static void
dump_selector(char *name, int selector)
{
    printk("%s: 0x%04X\n", name, selector);
    /* add routines to print limit... */
}

static void
print_frame_struct(struct int_frame *f)
{
    printk("eax= 0x%08X ebx= 0x%08X ecx= 0x%08X edx= 0x%08X\n",
	   f->eax, f->ebx, f->ecx, f->edx);
    printk("esi= 0x%08X edi= 0x%08X ebp= 0x%08X esp= 0x%08X\n",
	   f->esi, f->edi, f->ebp, f->esp);
    printk("eip= 0x%08X err= 0x%08X eflags= 0x%08X\n",
	   f->eip, f->err, f->eflags);
    dump_selector("cs", f->cs);
    dump_selector("ds", f->ds);
    dump_selector("es", f->es);
    dump_selector("fs", f->fs);
    dump_selector("gs", f->gs);
    dump_selector("ss", f->ss);
}

static page_faults = 0;

/* very simple exception handler (just prints some infos) */
void
handle_int(struct int_frame *f)
{
    if ((int_no >= 0x20) && (int_no < 0x30))
    {
	outb(0x20, 0x20);
	if (int_no >= 0x28)
	    outb(0xA0, 0x20);
	return;
    }

    if (int_no == 0x80)
    {
	syscall(f);
	f->eax = 0xDEADBEEF;
	return;
    }
    else
	printk("\n* interrupt %d *\n", int_no);

    print_frame_struct(f);

    if (int_no == 14) /* this is TEST CODE */
    {
	unsigned long address;

        asm("movl %%cr2, %0":"=p"(address));

	printk("page fault: eip=%p, address=%p, err=%p",
	       f->eip, address, f->err);

    } /* END OF TEST CODE */

    while (1);			/* should completly stop system instead of this */
}

