/*
 * LittleOS - http://come.to/littleos/
 * Copyright (C) 1998-2000 Lacroix Pascal <placr@hotmail.com>
 * All rights reserved.
 *
 * LittleOS is free software -- also known as "open source" -- you can
 * redistribute it and/or modify it under the terms of the GNU General
 * Public License, version 2, as published by the Free Software Foundation,
 * and found in the file `COPYING'.  This license basically allows free use,
 * modification, and redistribution, as long as the source to LittleOS and
 * any code linked against it is made freely available.  (The license in
 * `COPYING' is authoritative, not this interpretation of it.)
 *
 * LittleOS is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with LittleOS; see the file COPYING.  If not, write to the Free Software
 * Foundation, 59 Temple Place #330, Boston, MA 02111-1307, USA, or look
 * at http://www.fsf.org/copyleft/gpl.html.
 *
 */

/*
 * kernel_console.c: very simple console handling
 * Only simple console. Handles colors when you write "\27\[color]",
 * [color] is the desired color number (stdvga colors)
 *
 * For example: "\27\4Little\27\1OS" writes "Little" in red and "OS" in blue
 *
 */

#include <stdarg.h>
#include <string.h>
#include <sys/intr.h>
#include <i386/pio.h>

/* video memory pointer */
static short *screen = (short *) 0xB8000;

/* cursor position into video memory */
static int index;

/* current color (1..15) */
static int color = 7;

/* is console already initialized ? */
static int console_initialized = 0;

static void
update_cursor(void)
{
    int pos = index /* * 2 */ ;

    outb(0x3D4, 0xE);		/* select register */
    outb(0x3D5, (pos >> 8) & 0xFF);	/* data */
    outb(0x3D4, 0xF);
    outb(0x3D5, pos & 0xFF);
}

/* hascolor = 1 if the last character was '\27' */
static int hascolor;

void
putch(int c)
{
    if (hascolor)
    {
	if (c > 0 && c < 16)
	{
	    hascolor = 0;
	    color = c;
	}
	else
	    hascolor = 0;
    }
    else if (c == '\27')
	hascolor = 1;
    else if (c == '\n')
    {
	index = ((index + 80) / 80) * 80;	/* round up: new line */
    }
    else if (c != '\r')
    {
	screen[index++] = (color << 8) + c;
    }
}

int
puts(char *s)
{
    /* must we initialized the console ? */
    if (!console_initialized)
    {				/* yes */
	int i;

	console_initialized = 1;
	index = 0;
	hascolor = 0;
	for (i = 0; i < 80 * 25; i++)
	    screen[i] = (7 << 8) + ' ';
    }
    if (index >= 2000)		/* we have to scroll one line */
    {
	int i;

	index = 24 * 80;
	/* scroll screen */
	memcpy(screen, screen + 80, 80 * 24 * 2);
	/* clear last line */
	for (i = 24 * 80; i < 25 * 80; i++)
	    screen[i] = (7 << 8) + ' ';

    }
    /* print the text */
    while (*s)
	putch(*s++);

    update_cursor();

    return 1;
}

static char tmp_buf[512];
extern int vsprintf(char *, const char *, va_list);

/* this is the main kernel print routine with variable arguments */
int
printk(const char *format,...)
{
    int i;
    va_list args;

    va_start(args, format);
    i = vsprintf((char *) &tmp_buf, format, args);
    va_end(args);
    puts(tmp_buf);
    tmp_buf[0] = 0;
    return i;			/* return length of text */
}

void
panic(const char *format,...)
{
    va_list args;

    va_start(args, format);
    puts("\27\4kernel panic:\27\7 ");
    vsprintf((char *) &tmp_buf, format, args);
    va_end(args);
    puts(tmp_buf);
    disable_ints();
    while (1);
}

